<?php
/*
+-------------------------------------------------------------------------+
|                                                                         |
| @FILE_NAME:   Socket.abstract.php                                       |
| @CLASS_NAME:  (abstract) Jabber                                         |
|                                                                         |
| @DATE:        Wed May,16 2007 07:58:59                                  |
| @LAST_CHANGE: Wed May,16 2007 08:03:28                                  |
|                                                                         |
| @AUTHOR:      [NUTS] Rodrigo Souza - rodsouza@celepar.pr.gov.br         |
|                                                                         |
| @BRIEF:       Abstract class for connect with a socket                  |
|                                                                         |
+-------------------------------------------------------------------------+
*/

abstract class Socket
{
	protected final function open( $pSocket = false, $pTimeOut = 1, $pBlockingMode = 0 )
	{
      if ( !$pSocket )
         return false;

		$sock = socket_create(AF_INET, SOCK_STREAM, SOL_TCP);

      $pSocket = substr($pSocket, 6);
      if ( $socket = socket_connect($sock, substr($pSocket, 0, strpos($pSocket, ":")), substr($pSocket, strpos($pSocket, ":") + 1)) )
      {
         socket_set_block($sock);
      }

      return $sock;
	}

   protected final function blocking($pSocket = false, $pBlockingMode = 1)
   {
      if ( !$pSocket || !is_resource($pSocket) )
         return false;

      if ( $pBlockingMode == 1 )
      	socket_set_block($pSocket);
		else
			socket_set_nonblock($pSocket);
   }

	protected function read( $pSocket = false, $pLength = 4096 )
	{
      ini_set( 'track_errors', '1' );

      if ( !$pSocket || !is_resource($pSocket) )
         return false;

      $return = socket_read($pSocket, $pLength);

      if ( $php_errormsg )
         return false;

      return $return;
	}

	protected function write( $pSocket = false, $pData = false )
	{
      ini_set( 'track_errors', '1' );
      
      if ( (!$pSocket || !is_resource($pSocket)) || !$pData )
         return false;

		$return = @socket_write($pSocket, $pData);
      
      if ( $php_errormsg )
         return false;

      return $return;
	}

	protected final function close( $pSocket = false )
	{
      if ( !$pSocket || !is_resource($pSocket) )
         return false;

		return socket_close($pSocket);
	}
}
?>
